@@ -1,14 +1,12 @@
 BUILD_BIT='64bit'
 BUILD_CPU='Intel(R) Core(TM) i7-7700HQ CPU @ 2.80GHz'
-BUILD_DATE='2019-03-29'
+BUILD_DATE='2019-05-09'
 BUILD_MACHINE='x86_64'
 BUILD_OS='Linux   '
-BUILD_TIME='11:11'
-BUILT_BY='builduser'
-BUILT_ON='lock'
-COMPILER_VERSION='gcc (GCC) 8.2.1 20181127'
-CYTHON_VERSION='0.29.6'
-LINKER_VERSION='GNU ld (GNU Binutils) 2.31.1'
+BUILD_TIME='21:31'
+COMPILER_VERSION='gcc (GCC) 8.3.0'
+CYTHON_VERSION='0.29.7'
+LINKER_VERSION='GNU ld (GNU Binutils) 2.32'
 PYTHON_VERSION='2.7.16'
 RELEASE_BUILD=True
 lib_pycairo='1.18.0'
