@@ -1,14 +1,24 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2011-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2011-2019 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os.path
 import array
+
 from xpra.util import iround, first_time
-from xpra.os_util import strtobytes, bytestostr, WIN32, PYTHON2
-from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, import_glib, import_pixbufloader, import_pango, import_cairo, import_gobject, import_pixbuf, is_gtk3
+from xpra.os_util import strtobytes, bytestostr, WIN32, OSX, PYTHON2
+from xpra.gtk_common.gobject_compat import (
+    import_gtk, import_gdk, import_glib, import_pixbufloader,
+    import_pango, import_cairo, import_gobject, import_pixbuf, is_gtk3,
+    )
+from xpra.log import Logger
+
+log = Logger("gtk", "util")
+traylog = Logger("gtk", "tray")
+screenlog = Logger("gtk", "screen")
+
 gtk     = import_gtk()
 gdk     = import_gdk()
 pango   = import_pango()
@@ -18,12 +28,6 @@ glib    = import_glib()
 PixbufLoader = import_pixbufloader()
 Pixbuf = import_pixbuf()
 
-from xpra.log import Logger
-log = Logger("gtk", "util")
-traylog = Logger("gtk", "tray")
-screenlog = Logger("gtk", "screen")
-
-
 SHOW_ALL_VISUALS = False
 
 GTK_VERSION_INFO = {}
@@ -88,21 +92,57 @@ def popup_menu_workaround(*args):
     #only implemented with GTK2 on win32 below
     pass
 
+def pixbuf_save_to_memory(pixbuf, fmt="png"):
+    buf = []
+    def save_to_memory(data, *_args, **_kwargs):
+        buf.append(strtobytes(data))
+        return True
+    if is_gtk3():
+        pixbuf.save_to_callbackv(save_to_memory, None, fmt, [], [])
+    else:
+        pixbuf.save_to_callback(save_to_memory, fmt, {}, buf)
+    return b"".join(buf)
+
 
 if is_gtk3():
     def is_realized(widget):
         return widget.get_realized()
 
-    def make_temp_window(title):
+    def GDKWindow(parent=None, width=1, height=1, window_type=gdk.WindowType.TOPLEVEL,
+                  event_mask=0, wclass=gdk.WindowWindowClass.INPUT_OUTPUT, title=None,
+                  x=None, y=None, override_redirect=False, visual=None, **kwargs):
+        attributes_mask = 0
         attributes = gdk.WindowAttr()
-        attributes.width = 1
-        attributes.height = 1
-        attributes.title = title
-        attributes.window_type = gdk.WindowType.TOPLEVEL
-        attributes.wclass = gdk.WindowWindowClass.INPUT_OUTPUT
-        attributes.event_mask = 0
-        attributes_mask = gdk.WindowAttributesType.TITLE | gdk.WindowAttributesType.WMCLASS
-        return gdk.Window(None, attributes, attributes_mask)
+        if x is not None:
+            attributes.x = x
+            attributes_mask |= gdk.WindowAttributesType.X
+        if y is not None:
+            attributes.y = y
+            attributes_mask |= gdk.WindowAttributesType.Y
+        #attributes.type_hint = gdk.WindowTypeHint.NORMAL
+        #attributes_mask |= gdk.WindowAttributesType.TYPE_HINT
+        attributes.width = width
+        attributes.height = height
+        attributes.window_type = window_type
+        if title:
+            attributes.title = title
+            attributes_mask |= gdk.WindowAttributesType.TITLE
+        if visual:
+            attributes.visual = visual
+            attributes_mask |= gdk.WindowAttributesType.VISUAL
+        #OR:
+        attributes.override_redirect = override_redirect
+        attributes_mask |= gdk.WindowAttributesType.NOREDIR
+        #events:
+        attributes.event_mask = event_mask
+        #wclass:
+        attributes.wclass = wclass
+        mask = gdk.WindowAttributesType(attributes_mask)
+        return gdk.Window(parent, attributes, mask)
+
+    def make_temp_window(title):
+        return GDKWindow(title=title)
+
 
     def get_pixbuf_from_data(rgb_data, has_alpha, w, h, rowstride):
         data = array.array('B', strtobytes(rgb_data))
@@ -191,6 +231,7 @@ if is_gtk3():
     WIN_POS_CENTER  = gtk.WindowPosition.CENTER
     RESPONSE_CANCEL = gtk.ResponseType.CANCEL
     RESPONSE_OK     = gtk.ResponseType.OK
+    RESPONSE_ACCEPT = gtk.ResponseType.ACCEPT
 
     WINDOW_STATE_WITHDRAWN  = gdk.WindowState.WITHDRAWN
     WINDOW_STATE_ICONIFIED  = gdk.WindowState.ICONIFIED
@@ -202,10 +243,14 @@ if is_gtk3():
     FILE_CHOOSER_ACTION_SAVE    = gtk.FileChooserAction.SAVE
     FILE_CHOOSER_ACTION_OPEN    = gtk.FileChooserAction.OPEN
     PROPERTY_CHANGE_MASK = gdk.EventMask.PROPERTY_CHANGE_MASK
+    FOCUS_CHANGE_MASK    = gdk.EventMask.FOCUS_CHANGE_MASK
     BUTTON_PRESS_MASK    = gdk.EventMask.BUTTON_PRESS_MASK
     BUTTON_RELEASE_MASK  = gdk.EventMask.BUTTON_RELEASE_MASK
     ENTER_NOTIFY_MASK    = gdk.EventMask.ENTER_NOTIFY_MASK
     LEAVE_NOTIFY_MASK    = gdk.EventMask.LEAVE_NOTIFY_MASK
+    SUBSTRUCTURE_MASK    = gdk.EventMask.SUBSTRUCTURE_MASK
+    STRUCTURE_MASK       = gdk.EventMask.STRUCTURE_MASK
+    EXPOSURE_MASK        = gdk.EventMask.EXPOSURE_MASK
     ACCEL_LOCKED = gtk.AccelFlags.LOCKED
     ACCEL_VISIBLE = gtk.AccelFlags.VISIBLE
     JUSTIFY_LEFT    = gtk.Justification.LEFT
@@ -223,6 +268,14 @@ if is_gtk3():
     WINDOW_POPUP    = gtk.WindowType.POPUP
     WINDOW_TOPLEVEL = gtk.WindowType.TOPLEVEL
 
+    GDKWINDOW_TEMP     = gdk.WindowType.TEMP
+    GDKWINDOW_TOPLEVEL = gdk.WindowType.TOPLEVEL
+    GDKWINDOW_CHILD    = gdk.WindowType.CHILD
+    GDKWINDOW_FOREIGN  = gdk.WindowType.FOREIGN
+
+    CLASS_INPUT_OUTPUT = gdk.WindowWindowClass.INPUT_OUTPUT
+    CLASS_INPUT_ONLY = gdk.WindowWindowClass.INPUT_ONLY
+
     LSB_FIRST       = gdk.ByteOrder.LSB_FIRST
     MSB_FIRST       = gdk.ByteOrder.MSB_FIRST
     STATIC_GRAY     = gdk.VisualType.STATIC_GRAY
@@ -317,7 +370,8 @@ if is_gtk3():
     for x in ("PRIMARY", "SECONDARY", "CLIPBOARD"):
         CLIPBOARD_SELECTION[x] = getattr(gdk, "SELECTION_%s" % bytestostr(x))
     def GetClipboard(selection):
-        return Clipboard.get(CLIPBOARD_SELECTION[selection])
+        atom = CLIPBOARD_SELECTION.get(selection) or gdk.Atom.intern(bytestostr(selection), False)
+        return Clipboard.get(atom)
     def clipboard_request_contents(clipboard, target, unpack):
         target_atom = gdk.Atom.intern(bytestostr(target), False)
         clipboard.request_contents(target_atom, unpack)
@@ -364,8 +418,7 @@ if is_gtk3():
             return self.get_popup()
 
     gdk_window_process_all_updates = gdk.Window.process_all_updates
-    def gtk_main():
-        gtk.main()
+    gtk_main = gtk.main
 
     def gio_File(path):
         from gi.repository import Gio   #@UnresolvedImport
@@ -382,12 +435,27 @@ if is_gtk3():
         _, data, etag = gfile.load_contents_finish(res)
         return data, len(data), etag
 
+    def set_clipboard_data(clipboard, thevalue, _vtype="STRING"):
+        #only strings with GTK3?
+        thestring = str(thevalue)
+        clipboard.set_text(thestring, len(thestring))
+
+    def wait_for_contents(clipboard, target):
+        atom = gdk.Atom.intern(target, False)
+        return clipboard.wait_for_contents(atom)
+
+    PARAM_READABLE = gobject.ParamFlags.READABLE
+    PARAM_READWRITE = gobject.ParamFlags.READWRITE
 else:
     def get_pixbuf_from_data(rgb_data, has_alpha, w, h, rowstride):
         return gdk.pixbuf_new_from_data(rgb_data, gdk.COLORSPACE_RGB, has_alpha, 8, w, h, rowstride)
 
+    def GDKWindow(parent=None, width=1, height=1, window_type=gdk.WINDOW_TOPLEVEL,
+                  event_mask=0, wclass=gdk.INPUT_OUTPUT, title=None, x=-1, y=-1, **kwargs):
+        return gdk.Window(parent, width, height, window_type, event_mask, wclass, title, x, y, **kwargs)
+
     def make_temp_window(title):
-        return gdk.Window(None, 1, 1, gdk.WINDOW_TOPLEVEL, 0, gdk.INPUT_OUTPUT, title)
+        return GDKWindow(title=title)
 
     def get_pixbuf_from_window(window, x, y, w, h):
         pixbuf = gdk.Pixbuf(gdk.COLORSPACE_RGB, False, 8, w, h)
@@ -446,6 +514,7 @@ else:
     WIN_POS_CENTER  = gtk.WIN_POS_CENTER
     RESPONSE_CANCEL = gtk.RESPONSE_CANCEL
     RESPONSE_OK     = gtk.RESPONSE_OK
+    RESPONSE_ACCEPT = gtk.RESPONSE_ACCEPT
 
     WINDOW_STATE_WITHDRAWN  = gdk.WINDOW_STATE_WITHDRAWN
     WINDOW_STATE_ICONIFIED  = gdk.WINDOW_STATE_ICONIFIED
@@ -457,10 +526,14 @@ else:
     FILE_CHOOSER_ACTION_SAVE    = gtk.FILE_CHOOSER_ACTION_SAVE
     FILE_CHOOSER_ACTION_OPEN    = gtk.FILE_CHOOSER_ACTION_OPEN
     PROPERTY_CHANGE_MASK = gdk.PROPERTY_CHANGE_MASK
+    FOCUS_CHANGE_MASK    = gdk.FOCUS_CHANGE_MASK
     BUTTON_PRESS_MASK    = gdk.BUTTON_PRESS_MASK
     BUTTON_RELEASE_MASK  = gdk.BUTTON_RELEASE_MASK
     ENTER_NOTIFY_MASK    = gdk.ENTER_NOTIFY_MASK
     LEAVE_NOTIFY_MASK    = gdk.LEAVE_NOTIFY_MASK
+    SUBSTRUCTURE_MASK    = gdk.SUBSTRUCTURE_MASK
+    STRUCTURE_MASK       = gdk.STRUCTURE_MASK
+    EXPOSURE_MASK        = gdk.EXPOSURE_MASK
     ACCEL_LOCKED = gtk.ACCEL_LOCKED
     ACCEL_VISIBLE = gtk.ACCEL_VISIBLE
     JUSTIFY_LEFT    = gtk.JUSTIFY_LEFT
@@ -478,6 +551,14 @@ else:
     WINDOW_POPUP    = gtk.WINDOW_POPUP
     WINDOW_TOPLEVEL = gtk.WINDOW_TOPLEVEL
 
+    GDKWINDOW_TEMP     = gdk.WINDOW_TEMP
+    GDKWINDOW_TOPLEVEL = gdk.WINDOW_TOPLEVEL
+    GDKWINDOW_CHILD    = gdk.WINDOW_CHILD
+    GDKWINDOW_FOREIGN  = gdk.WINDOW_FOREIGN
+
+    CLASS_INPUT_OUTPUT = gdk.INPUT_OUTPUT
+    CLASS_INPUT_ONLY = gdk.INPUT_ONLY
+
     LSB_FIRST       = gdk.LSB_FIRST
     MSB_FIRST       = gdk.MSB_FIRST
     STATIC_GRAY     = gdk.VISUAL_STATIC_GRAY
@@ -621,10 +702,11 @@ else:
             mouse_in_tray_menu = False
             def check_menu_left(expected_counter):
                 if mouse_in_tray_menu:
-                    return    False
+                    return False
                 if expected_counter!=mouse_in_tray_menu_counter:
-                    return    False            #counter has changed
+                    return False            #counter has changed
                 close_cb()
+                return False
             gobject.timeout_add(OUTSIDE_TRAY_TIMEOUT, check_menu_left, mouse_in_tray_menu_counter)
         mouse_in_tray_menu_counter = 0
         mouse_in_tray_menu = False
@@ -645,12 +727,39 @@ else:
     def load_contents_finish(gfile, res):
         return gfile.load_contents_finish(res)
 
+    def set_clipboard_data(clipboard, thevalue, vtype="STRING"):
+        TEXT_TARGETS = ("UTF8_STRING", "TEXT", "STRING", "text/plain")
+        if WIN32:
+            if vtype in TEXT_TARGETS:
+                clipboard.set_text(thevalue, len(thevalue))
+            return
+        if OSX and vtype in TEXT_TARGETS:
+            clipboard.set_text(thevalue, len(thevalue))
+            return
+        value = [thevalue]
+        def get_func(clipboard, selection, info, targets):
+            log("get_func%s value=%s", (clipboard, selection, info, targets), value[0])
+            s = value[0]
+            if s:
+                selection.set(vtype, 8, s)
+            else:
+                clipboard.clear()
+        def clear_func(*args):
+            log("clear_func%s value=%s", args, value[0])
+            value[0] = ""
+        clipboard.set_with_data([(vtype, 0, 0)], get_func, clear_func)
+
+    def wait_for_contents(clipboard, target):
+        return clipboard.wait_for_contents(target)
+
+    PARAM_READABLE = gobject.PARAM_READABLE
+    PARAM_READWRITE = gobject.PARAM_READWRITE
+
 #no idea why, but trying to use the threads_init / threads_enter
 #causes deadlocks on win32:
 if WIN32:
-    def gtk_main():
-        gtk.main()
-
+    gtk_main = gtk.main
+gtk_main_quit = gtk.main_quit
 
 GRAB_STATUS_STRING = {
                       GRAB_SUCCESS          : "SUCCESS",
@@ -694,7 +803,8 @@ class TrayCheckMenuItem(gtk.CheckMenuItem):
         traylog("TrayCheckMenuItem.on_button_release_event(%s) label=%s", args, self.label)
         self.active_state = self.get_active()
         def recheck():
-            traylog("TrayCheckMenuItem: recheck() active_state=%s, get_active()=%s", self.active_state, self.get_active())
+            traylog("TrayCheckMenuItem: recheck() active_state=%s, get_active()=%s",
+                    self.active_state, self.get_active())
             state = self.active_state
             self.active_state = None
             if state is not None and state==self.get_active():
@@ -733,9 +843,9 @@ class TrayImageMenuItem(gtk.ImageMenuItem):
 
 ImageMenuItemClass = gtk.ImageMenuItem
 CheckMenuItemClass = gtk.CheckMenuItem
-def CheckMenuItem(*args, **kwargs):
+def CheckMenuItem(label):
     global CheckMenuItemClass
-    return CheckMenuItemClass(*args, **kwargs)
+    return CheckMenuItemClass(label=label)
 
 def set_use_tray_workaround(enabled):
     global CheckMenuItemClass, ImageMenuItemClass
@@ -888,11 +998,32 @@ def get_screen_info(display, screen):
         if fo:
             fontoptions = info.setdefault("fontoptions", {})
             for x,vdict in {
-                            "antialias"     : {cairo.ANTIALIAS_DEFAULT      : "default", cairo.ANTIALIAS_NONE       : "none",   cairo.ANTIALIAS_GRAY        : "gray",   cairo.ANTIALIAS_SUBPIXEL    : "subpixel"},
-                            "hint_metrics"  : {cairo.HINT_METRICS_DEFAULT   : "default", cairo.HINT_METRICS_OFF     : "off",    cairo.HINT_METRICS_ON       : "on"},
-                            "hint_style"    : {cairo.HINT_STYLE_DEFAULT     : "default", cairo.HINT_STYLE_NONE      : "none",   cairo.HINT_STYLE_SLIGHT     : "slight", cairo.HINT_STYLE_MEDIUM     : "medium", cairo.HINT_STYLE_FULL       : "full"},
-                            "subpixel_order": {cairo.SUBPIXEL_ORDER_DEFAULT : "default", cairo.SUBPIXEL_ORDER_RGB   : "RGB",    cairo.SUBPIXEL_ORDER_BGR    : "BGR",    cairo.SUBPIXEL_ORDER_VRGB   : "VRGB",   cairo.SUBPIXEL_ORDER_VBGR   : "VBGR"},
-                            }.items():
+                "antialias" : {
+                    cairo.ANTIALIAS_DEFAULT     : "default",
+                    cairo.ANTIALIAS_NONE        : "none",
+                    cairo.ANTIALIAS_GRAY        : "gray",
+                    cairo.ANTIALIAS_SUBPIXEL    : "subpixel",
+                    },
+                "hint_metrics" : {
+                    cairo.HINT_METRICS_DEFAULT  : "default",
+                    cairo.HINT_METRICS_OFF      : "off",
+                    cairo.HINT_METRICS_ON       : "on",
+                    },
+                "hint_style" : {
+                    cairo.HINT_STYLE_DEFAULT    : "default",
+                    cairo.HINT_STYLE_NONE       : "none",
+                    cairo.HINT_STYLE_SLIGHT     : "slight",
+                    cairo.HINT_STYLE_MEDIUM     : "medium",
+                    cairo.HINT_STYLE_FULL       : "full",
+                    },
+                "subpixel_order": {
+                    cairo.SUBPIXEL_ORDER_DEFAULT    : "default",
+                    cairo.SUBPIXEL_ORDER_RGB        : "RGB",
+                    cairo.SUBPIXEL_ORDER_BGR        : "BGR",
+                    cairo.SUBPIXEL_ORDER_VRGB       : "VRGB",
+                    cairo.SUBPIXEL_ORDER_VBGR       : "VBGR",
+                    },
+                }.items():
                 fn = getattr(fo, "get_"+x)
                 val = fn()
                 fontoptions[x] = vdict.get(val, val)
@@ -902,24 +1033,37 @@ def get_screen_info(display, screen):
     def visual(name, v):
         if not v:
             return
-        for x, vdict in {"bits_per_rgb" : {},
-                         "byte_order"   : {LSB_FIRST    : "LSB", MSB_FIRST  : "MSB"},
-                         "colormap_size": {},
-                         "depth"        : {},
-                         "red_pixel_details"    : {},
-                         "green_pixel_details"  : {},
-                         "blue_pixel_details"   : {},
-                         "visual_type"  : {STATIC_GRAY : "STATIC_GRAY", GRAYSCALE : "GRAYSCALE",  STATIC_COLOR : "STATIC_COLOR", PSEUDO_COLOR : "PSEUDO_COLOR", TRUE_COLOR : "TRUE_COLOR", DIRECT_COLOR : "DIRECT_COLOR"},
-                         }.items():
+        for x, vdict in {
+            "bits_per_rgb"          : {},
+            "byte_order"            : {
+                LSB_FIRST : "LSB",
+                MSB_FIRST : "MSB",
+            },
+            "colormap_size"         : {},
+            "depth"                 : {},
+            "red_pixel_details"     : {},
+            "green_pixel_details"   : {},
+            "blue_pixel_details"    : {},
+            "visual_type"           : {
+                STATIC_GRAY    : "STATIC_GRAY",
+                GRAYSCALE      : "GRAYSCALE",
+                STATIC_COLOR   : "STATIC_COLOR",
+                PSEUDO_COLOR   : "PSEUDO_COLOR",
+                TRUE_COLOR     : "TRUE_COLOR",
+                DIRECT_COLOR   : "DIRECT_COLOR",
+                },
+            }.items():
             val = None
             try:
-                val = getattr(v, x.replace("visual_"))  #ugly workaround for "visual_type" -> "type" for GTK2...
-            except:
+                #ugly workaround for "visual_type" -> "type" for GTK2...
+                val = getattr(v, x.replace("visual_"))
+            except AttributeError:
                 try:
                     fn = getattr(v, "get_"+x)
-                    val = fn()
-                except:
+                except AttributeError:
                     pass
+                else:
+                    val = fn()
             if val is not None:
                 vinfo.setdefault(name, {})[x] = vdict.get(val, val)
     try:
@@ -1111,7 +1255,8 @@ def set_tooltip_text(widget, text):
         widget.set_tooltip_text(text)
 
 
-def imagebutton(title, icon, tooltip=None, clicked_callback=None, icon_size=32, default=False, min_size=None, label_color=None, label_font=None):
+def imagebutton(title, icon, tooltip=None, clicked_callback=None, icon_size=32,
+                default=False, min_size=None, label_color=None, label_font=None):
     button = gtk.Button(title)
     settings = button.get_settings()
     settings.set_property('gtk-button-images', True)
@@ -1135,7 +1280,7 @@ def imagebutton(title, icon, tooltip=None, clicked_callback=None, icon_size=32,
             alignment = button.get_children()[0]
             b_hbox = alignment.get_children()[0]
             label = b_hbox.get_children()[1]
-        except:
+        except IndexError:
             pass
         else:
             if label_color:
@@ -1217,7 +1362,7 @@ def window_defaults(window):
 #selected in a submenu:
 def ensure_item_selected(submenu, item, recurse=True):
     if not isinstance(item, gtk.CheckMenuItem):
-        return
+        return None
     if item.get_active():
         #deactivate all except this one
         def deactivate(items, skip=None):
@@ -1248,7 +1393,7 @@ def ensure_item_selected(submenu, item, recurse=True):
         return None
     active = get_active_item(submenu.get_children())
     if active:
-        return  active
+        return active
     #if not then keep this one active:
     item.set_active(True)
     return item
@@ -1282,7 +1427,8 @@ class TableBuilder(object):
         self.inc()
 
     def attach(self, widget, i, count=1, xoptions=FILL, yoptions=FILL, xpadding=10, ypadding=0):
-        self.table.attach(widget, i, i+count, self.row, self.row+1, xoptions=xoptions, yoptions=yoptions, xpadding=xpadding, ypadding=ypadding)
+        self.table.attach(widget, i, i+count, self.row, self.row+1,
+                          xoptions=xoptions, yoptions=yoptions, xpadding=xpadding, ypadding=ypadding)
 
     def inc(self):
         self.row += 1
@@ -1306,7 +1452,7 @@ def choose_file(parent_window, title, action, action_button, callback=None, file
     chooser.hide()
     chooser.destroy()
     if response!=RESPONSE_OK or len(filenames)!=1:
-        return
+        return None
     filename = filenames[0]
     if callback:
         callback(filename)
